package thesis.gui;

import java.awt.AWTEvent;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import thesis.tree.Vector;

public class Universe extends JComponent implements ChangeListener {

  private static final long serialVersionUID = 1L;
  
  public static final int WIDTH = 560;
  public static final int HEIGHT = 315;
  
  private List<Galaxy> galaxies;
  private SpinnerNumberModel galaxyIndex;
  
  public static AffineTransform getTransform() {
    AffineTransform transform = new AffineTransform();
    transform.scale(WIDTH / 16, HEIGHT / 9);
    transform.translate(8.0, 4.5);
    return transform;
  }
  
  public static AffineTransform getInverseTransform() {
    AffineTransform transform = new AffineTransform();
    transform.translate(-8.0, -4.5);
    transform.scale(16.0 / WIDTH, 9.0 / HEIGHT);
    return transform;
  }
  
  public Universe(List<Galaxy> galaxies, SpinnerNumberModel galaxyIndex) {
    this.galaxies = galaxies;
    this.galaxyIndex = galaxyIndex;
    
    galaxyIndex.addChangeListener(this);
    enableEvents(AWTEvent.MOUSE_EVENT_MASK | AWTEvent.MOUSE_MOTION_EVENT_MASK);
  }
  
  private void setSpinnerValue(SpinnerNumberModel model, double value) {
    if (value < (Double)model.getMinimum()) {
      model.setValue(model.getMinimum());
    } else if (value > (Double)model.getMaximum()) {
      model.setValue(model.getMaximum());
    } else {
      model.setValue(value);
    }
  }
  
  private void changeGalaxyPosition(int x, int y) {
    Galaxy galaxy = galaxies.get(galaxyIndex.getNumber().intValue() - 1);
    Point2D point = getInverseTransform().transform(
        new Point2D.Float(x, y), null);
    
    setSpinnerValue(galaxy.getPositionX(), point.getX());
    setSpinnerValue(galaxy.getPositionY(), point.getY());
    
    repaint();
  }
  
  private void changeGalaxyVelocity(int x, int y) {
    Galaxy galaxy = galaxies.get(galaxyIndex.getNumber().intValue() - 1);
    Point2D point = getInverseTransform().transform(
        new Point2D.Float(x, y), null);
    
    setSpinnerValue(galaxy.getVelocityX(),
        point.getX() - galaxy.getPositionX().getNumber().floatValue());
    
    setSpinnerValue(galaxy.getVelocityY(),
        point.getY() - galaxy.getPositionY().getNumber().floatValue());
    
    repaint();
  }
  
  @Override
  public void stateChanged(ChangeEvent e) {
    repaint();
  }
  
  @Override
  protected void processMouseEvent(MouseEvent e) {
    super.processMouseEvent(e);
    
    if (e.getID() == MouseEvent.MOUSE_PRESSED) {
      if ((e.getModifiersEx() & MouseEvent.BUTTON1_DOWN_MASK) != 0) {
        requestFocusInWindow();
        changeGalaxyPosition(e.getX(), e.getY());
      } else if ((e.getModifiersEx() & MouseEvent.BUTTON3_DOWN_MASK) != 0) {
        requestFocusInWindow();
        changeGalaxyVelocity(e.getX(), e.getY());
      }
    }
  }
  
  @Override
  protected void processMouseMotionEvent(MouseEvent e) {
    super.processMouseMotionEvent(e);
    
    if (e.getID() == MouseEvent.MOUSE_DRAGGED) {
      if ((e.getModifiersEx() & MouseEvent.BUTTON1_DOWN_MASK) != 0) {
        changeGalaxyPosition(e.getX(), e.getY());
      } else if ((e.getModifiersEx() & MouseEvent.BUTTON3_DOWN_MASK) != 0) {
        changeGalaxyVelocity(e.getX(), e.getY());
      }
    }
  }
  
  @Override
  public Dimension getMinimumSize() {
    return getPreferredSize();
  }
  
  @Override
  public Dimension getPreferredSize() {
    return new Dimension(WIDTH, HEIGHT);
  }
  
  @Override
  public Dimension getMaximumSize() {
    return getPreferredSize();
  }
  
  @Override
  public void paintComponent(Graphics graphics) {
    super.paintComponent(graphics);
    Graphics2D g = (Graphics2D)graphics;
    g.setColor(Color.BLACK);
    g.fillRect(0, 0, getWidth(), getHeight());
    
    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
        RenderingHints.VALUE_ANTIALIAS_ON);
    
    g.setStroke(
        new BasicStroke(0.15f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_MITER));
    
    Galaxy current = galaxies.get(galaxyIndex.getNumber().intValue() - 1);
    
    for (int i = 0; i < galaxies.size(); ++i) {
      Galaxy galaxy = galaxies.get(i);
      if (galaxy != current) {
        drawGalaxy(g, galaxy, Color.GRAY, new Color(0, 0, 128));
      }
    }
    drawGalaxy(g, current, Color.WHITE, Color.BLUE);
  }
  
  private void drawGalaxy(Graphics2D g, Galaxy galaxy, Color galaxyColor,
      Color velocityColor) {
    
    final float ARROW_WIDTH = 0.4f;
    final float ARROW_LENGTH = 1.0f;
    
    AffineTransform oldTransform = g.getTransform();
    g.transform(getTransform());

    float x = galaxy.getPositionX().getNumber().floatValue();
    float y = galaxy.getPositionY().getNumber().floatValue();
    float r = galaxy.getRadius().getNumber().floatValue();
    g.translate(x, y);
    
    g.setColor(galaxyColor);
    g.fill(new Ellipse2D.Float(-r, -r, 2 * r, 2 * r));
    
    float vx = galaxy.getVelocityX().getNumber().floatValue();
    float vy = galaxy.getVelocityY().getNumber().floatValue();
    Vector velocity = new Vector(vx, vy);
    g.rotate(velocity.angle() - Math.PI / 2);
    g.setColor(velocityColor);
    
    if (velocity.length() < ARROW_LENGTH) {
      GeneralPath triangle = new GeneralPath();
      triangle.moveTo(-ARROW_WIDTH / 2, 0);
      triangle.lineTo(ARROW_WIDTH / 2, 0);
      triangle.lineTo(0, velocity.length());
      triangle.closePath();
      g.fill(triangle);
    } else {
      g.draw(new Line2D.Float(0, 0, 0, velocity.length() - ARROW_LENGTH));
      g.translate(0, velocity.length() - ARROW_LENGTH);
      
      GeneralPath triangle = new GeneralPath();
      triangle.moveTo(-ARROW_WIDTH / 2, 0);
      triangle.lineTo(ARROW_WIDTH / 2, 0);
      triangle.lineTo(0, ARROW_LENGTH);
      triangle.closePath();
      g.fill(triangle);
    }
    
    g.setTransform(oldTransform);
  }

}
