package thesis.gui;

import java.awt.FileDialog;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipFile;

import javax.imageio.ImageIO;
import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextField;
import javax.swing.Timer;
import javax.swing.GroupLayout.Alignment;
import javax.swing.LayoutStyle.ComponentPlacement;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import thesis.gui.icons.Icons;

public final class VideoDialog extends JDialog {

  private static final long serialVersionUID = 1L;
  
  private ZipFile videoFile;
  private Timer timer;
  
  private JButton playButton;
  private JButton pauseButton;
  private JSlider position;
  private ImagePanel video;
  
  public VideoDialog(Window owner) {
    super(owner, "Video", DEFAULT_MODALITY_TYPE);
    setDefaultCloseOperation(DISPOSE_ON_CLOSE);
    
    addWindowListener(new WindowAdapter() {
      @Override
      public void windowClosed(WindowEvent e) {
        timer.stop();
        closeVideoFile();
      }
    });
    
    this.timer = new Timer(100, new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        if (position.getValue() != position.getMaximum()) {
          position.setValue(position.getValue() + 1);
        } else {
          pauseVideo();
        }
      }
    });

    setResizable(false);
    setContentPane(createContentPanel());
    pack();
    setLocationRelativeTo(owner);
  }
  
  
  
  private void loadVideo(String filename) {
    if (pauseButton.isVisible()) {
      pauseVideo();
    }
    closeVideoFile();
    
    try {
      videoFile = new ZipFile(filename);
    } catch (Exception e) {
      Application.showError(
          "Dogodila se greška tijekom čitanja video snimke.");
      return;
    }
    
    playButton.setEnabled(true);
    position.setEnabled(true);
    position.setValue(0);
    position.setMaximum(videoFile.size() - 1);
  }
  
  private void closeVideoFile() {
    if (videoFile != null) {
      try {
        videoFile.close();
      } catch (IOException ignored) {
        
      }
      
      videoFile = null;
    }
  }
  
  private void playVideo() {
    pauseButton.setVisible(true);
    pauseButton.requestFocusInWindow();
    playButton.setVisible(false);
    
    if (position.getValue() == position.getMaximum()) {
      position.setValue(0);
    }
    timer.start();
  }
  
  private void pauseVideo() {
    timer.stop();
    playButton.setVisible(true);
    if (pauseButton.hasFocus()) {
      playButton.requestFocusInWindow();
    }
    pauseButton.setVisible(false);
    
  }
  
  private void showFrame() {
    int index = position.getValue();
    StringBuilder filename = new StringBuilder(Integer.toString(index));
    while (filename.length() < 8) {
      filename.insert(0, '0');
    }
    
    try {
      InputStream input = videoFile.getInputStream(
          videoFile.getEntry(filename.append(".png").toString()));
      video.setImage(ImageIO.read(input));
      video.repaint();
      input.close();
    } catch (Exception e) {
      Application.showError("Dogodila se greška tijekom čitanja video snimke.");
      
      if (pauseButton.isVisible()) {
        pauseVideo();
      }
      playButton.setEnabled(false);
      position.setEnabled(false);
      video.setImage(null);
    }
  }

  private JPanel createContentPanel() {
    JLabel fileLabel = new JLabel("Datoteka:");
    final JTextField fileTextField = new JTextField();
    JButton fileButton = new JButton("...");
    
    fileTextField.setEditable(false);
    fileButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        if (Application.openFileDialog(FileDialog.LOAD, fileTextField)) {
          loadVideo(fileTextField.getText());
        }
      }
    });
    
    playButton = new JButton(Icons.load("play.png"));
    pauseButton = new JButton(Icons.load("pause.png"));
    
    playButton.setEnabled(false);
    playButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        playVideo();
      }
    });
    
    pauseButton.setVisible(false);
    pauseButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        pauseVideo();
      }
    });
    
    position = new JSlider(0, 1000);
    position.setEnabled(false);
    position.setValue(0);
    position.setMajorTickSpacing(1);
    position.setSnapToTicks(true);
    position.addChangeListener(new ChangeListener() {
      @Override
      public void stateChanged(ChangeEvent e) {
        showFrame();
      }
    });
    
    video = new ImagePanel(Universe.WIDTH, Universe.HEIGHT);
    
    JPanel panel = new JPanel();
    GroupLayout layout = new GroupLayout(panel);
    panel.setLayout(layout);
    
    layout.setHorizontalGroup(layout.createSequentialGroup()
      .addContainerGap()
      .addGroup(layout.createParallelGroup()
        .addGroup(layout.createSequentialGroup()
          .addComponent(fileLabel)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(fileTextField)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(fileButton)
        )
        .addGroup(layout.createSequentialGroup()
          .addComponent(playButton)
          .addComponent(pauseButton)
          .addPreferredGap(ComponentPlacement.RELATED)
          .addComponent(position)
        )
        .addComponent(video)
      )
      .addContainerGap()
    );
    
    layout.setVerticalGroup(layout.createSequentialGroup()
      .addContainerGap()
      .addGroup(layout.createParallelGroup(Alignment.BASELINE)
        .addComponent(fileLabel)
        .addComponent(fileTextField)
        .addComponent(fileButton)
      )
      .addPreferredGap(ComponentPlacement.UNRELATED)
      .addGroup(layout.createParallelGroup(Alignment.CENTER)
        .addComponent(playButton)
        .addComponent(pauseButton)
        .addComponent(position)
      )
      .addPreferredGap(ComponentPlacement.UNRELATED)
      .addComponent(video)
      .addContainerGap()
    );
    
    return panel;
  }

}
