package thesis.simulation;

import java.util.List;

import javax.swing.SwingUtilities;
import javax.swing.event.ChangeListener;

import thesis.tree.Body;
import thesis.tree.TreeNode;
import thesis.tree.Vector;

public abstract class Simulator extends Thread {
  
  public enum Timer {
    
    TREE_CONSTRUCTION,
    TREE_PROCESSING,
    FORCE_CALCULATION,
    FORCE_INTEGRATION,
    VIDEO_RECORDING;
    
  }
  
  private ChangeListener listener;
  
  protected List<Body> bodies;
  private int steps;
  
  private VideoRecorder recorder;
  
  protected Integrator integrator;
  protected TreeNode root;
  protected float minX, minY;
  protected float maxX, maxY;
  
  private long[] timers;

  public Simulator(List<Body> bodies, int steps) {
    this.bodies = bodies;
    this.steps = steps;
    this.timers = new long[Timer.values().length];
    
    minX = minY = Float.POSITIVE_INFINITY;
    maxX = maxY = Float.NEGATIVE_INFINITY;
    for (Body body : bodies) {
      adjustBounds(body);
    }
  }

  private void adjustBounds(Body body) {
    minX = Math.min(minX, body.getPosition().getX());
    minY = Math.min(minY, body.getPosition().getY());
    maxX = Math.max(maxX, body.getPosition().getX());
    maxY = Math.max(maxY, body.getPosition().getY());
  }
  
  @Override
  public final void run() {
    try {
      for (int i = 0; i < steps; ++i) {
        buildTree();
        processTree();
        calculateForces();
        integrate();
        recordVideo();
        fireChangeEvent();
      }
    } catch (InterruptedException e) {
      // simulation aborted - quit
    }
    
    cleanup();
  }

  private void fireChangeEvent() {
    if (listener != null) {
      SwingUtilities.invokeLater(new Runnable() {
        @Override
        public void run() {
          listener.stateChanged(null);
        }
      });
    }
  }

  private void buildTree()  throws InterruptedException {
    long time = System.currentTimeMillis();
    
    Vector center = new Vector((minX + maxX) / 2, (minY + maxY) / 2);
    float size = Math.max(maxX - minX, maxY - minY) / 2;
    
    root = new TreeNode(center, size, bodies.get(0));
    for (int i = 1; i < bodies.size(); ++i) {
      if (Thread.interrupted()) {
        throw new InterruptedException();
      }
      
      root.insert(bodies.get(i));
    }
    
    timers[Timer.TREE_CONSTRUCTION.ordinal()] +=
      System.currentTimeMillis() - time;
  }
  
  private void processTree() throws InterruptedException {
    long time = System.currentTimeMillis();
    
    root.calculateCenterOfMass();
    
    timers[Timer.TREE_PROCESSING.ordinal()] +=
      System.currentTimeMillis() - time;
  }
  
  private void calculateForces() throws InterruptedException {
    long time = System.currentTimeMillis();

    doCalculateForces();
    
    timers[Timer.FORCE_CALCULATION.ordinal()] +=
      System.currentTimeMillis() - time;
  }
  
  protected abstract void doCalculateForces() throws InterruptedException;
  
  private void integrate() throws InterruptedException {
    long time = System.currentTimeMillis();
    
    for (Body body : bodies) {
      integrator.integrate(body);
      adjustBounds(body);
    }
    
    timers[Timer.FORCE_INTEGRATION.ordinal()] +=
      System.currentTimeMillis() - time;
  }
  
  private void recordVideo() {
    long time = System.currentTimeMillis();
    
    if (recorder != null) {
      recorder.writeFrame(bodies);
    }
    
    timers[Timer.VIDEO_RECORDING.ordinal()] +=
      System.currentTimeMillis() - time;
  }
  
  protected void cleanup() {
    if (recorder != null) {
      recorder.close();
    }
  }
  
  public void setChangeListener(ChangeListener listener) {
    this.listener = listener;
  }
  
  public void setIntegrator(Integrator integrator) {
    this.integrator = integrator;
  }
  
  public void setRecorder(VideoRecorder recorder) {
    this.recorder = recorder;
  }
  
  public long getTimer(int index) {
    return timers[index];
  }
  
}
